#!/usr/bin/python
import time, sys, os, getopt
import struct, array
import serial
import crcmod
import socket

# Sizes
ROW_SIZE = 512
PAGE_SIZE = 4096

# States
STATE_IDLE = 0
STATE_HEADER = 1
STATE_PAYLOAD = 2

# Packet types
TYPE_CONNECT = 0
TYPE_WRITE_ROW = 1
TYPE_ERASE_PAGE = 2
TYPE_START_APP = 3
TYPE_CONNECT_ACK = TYPE_CONNECT | 0x80
TYPE_WRITE_ROW_ACK = TYPE_WRITE_ROW | 0x80
TYPE_ERASE_PAGE_ACK = TYPE_ERASE_PAGE | 0x80
TYPE_START_APP_ACK = TYPE_START_APP | 0x80

# Constants
START_BYTE = '\x24'
COM_TIMEOUT = 1.2   # Should be at least 1 seconds to allow PIC to write row and BL com-timeout is 1.0
CONNECT_RETRIES = 10
GENERAL_RETRIES = 5
SERIAL_BAUD = 115200

# UDP related settings
UDP_TX_PORT = 20011
UDP_RX_IP = "225.0.0.37" # Multicast group
UDP_RX_PORT = 20001      # Multicast port
MULTICAST_SCAN_TIME = 1  # In seconds

class bootloader:
    def __init__(self):
        self.state = STATE_IDLE
        self.pl = ""
        self.crc = 0
        self.type = 0
        self.len = 0
        self.crc_func = crcmod.mkCrcFun(0x104c11db7, initCrc=0, xorOut=0xffffffff)
        self.crc_calc = 0
        self.connected = False
        self.file_path = False
        self.total_rows = 0
        self.total_pages = 0
        self.ser = False    # Pyserial instance
        self.rx_skt = False # UDP multicast reception socket
        self.tx_skt = False # UDP transmit socket
        self.target_ip = False
        
    def __del__(self):
        self.close()
        
    def close(self):
        if self.ser:
            try:
                self.ser.close()
                self.ser = False
            except Exception:
                pass
        if self.rx_skt:
            self.rx_skt.close()
        if self.tx_skt:
            self.tx_skt.close()

    def serial_open(self, portname, portbaud):
        try:
            self.ser = serial.Serial(portname,portbaud,timeout=0)
            return True
        except Exception:
            return False

    def serial_scan(self):
        if sys.platform.startswith('win'):
            ports = ['COM' + str(i + 1) for i in range(256)]
        elif sys.platform.startswith('linux'):
            import glob
            ports = glob.glob('/dev/tty[A-Za-z]*')
        else:
            raise EnvironmentError('Unsupported platform')

        result = []
        for port in ports:
            try:
                s = serial.Serial(port)
                s.close()
                result.append(port)
            except (OSError, serial.SerialException):
                pass
        return result

    def socket_open(self, ip):
        self.target_ip = ip
        self.tx_skt = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # UDP TX socket
        if not self.rx_skt:
            self.rx_skt = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            self.rx_skt.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.rx_skt.setblocking(0)
            self.rx_skt.bind(('', UDP_RX_PORT))
            mreq = struct.pack("4sl", socket.inet_aton(UDP_RX_IP), socket.INADDR_ANY)
            self.rx_skt.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        return True

    def socket_scan(self):
        addresses = []
        start_time = time.time()

        if not self.rx_skt:
            self.rx_skt = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            self.rx_skt.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.rx_skt.setblocking(0)
            self.rx_skt.bind(('', UDP_RX_PORT))
            mreq = struct.pack("4sl", socket.inet_aton(UDP_RX_IP), socket.INADDR_ANY)
            self.rx_skt.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

        while ((time.time()-start_time) < MULTICAST_SCAN_TIME ):
            try:
                data,address = self.rx_skt.recvfrom(1000)
            except socket.error, e:
                pass
            else:
                if not (address[0] in addresses):
                    addresses.append(address[0])
            time.sleep(0.1)
        return addresses

    def write(self, data):
        if self.ser:
            self.ser.write(data)
        elif self.tx_skt and self.target_ip:
            self.tx_skt.sendto(data, (self.target_ip, UDP_TX_PORT))
        else:
            return False

    def read(self):
        if self.ser:
            return self.ser.read(100)
        elif self.rx_skt:
            try:
                data,address = self.rx_skt.recvfrom(1000)
            except socket.error:
                pass
            else:
                if address[0] == self.target_ip:
                    return data
        else:
            return ""
            
    def erase_page(self, page_nr):
        crc = self.crc_func(struct.pack('<BH', TYPE_ERASE_PAGE, page_nr))
        pkt = struct.pack('<BIBH', ord(START_BYTE), crc, TYPE_ERASE_PAGE, page_nr)
        
        # Wait for connection with retry
        for j in range(0,GENERAL_RETRIES):
            self.write(pkt)
            # Continue to receive until desired packet is received (or timeout)
            while (self.wait_pkt()):
                if (self.type == TYPE_ERASE_PAGE_ACK):
                    [nr, ack] = struct.unpack("<HB", self.pl)
                    if (ack and (nr == page_nr)):
                        return True
        return False

    def write_row(self, row_nr, row_data):
        if (len(row_data) == 512):
            crc = self.crc_func(struct.pack('<BH', TYPE_WRITE_ROW, row_nr) + row_data)
            pkt = struct.pack('<BIBH', ord(START_BYTE), crc, TYPE_WRITE_ROW, row_nr) + row_data
            
            # Wait for connection with retry
            for j in range(0,GENERAL_RETRIES):
                self.write(pkt)
                # Continue to receive until desired packet is received (or timeout)
                while (self.wait_pkt()):
                    if (self.type == TYPE_WRITE_ROW_ACK):
                        [nr, ack] = struct.unpack("<HB", self.pl)
                        if (ack and (nr == row_nr)):
                            return True
        return False

    def start_app(self):
        # Wait for ACK (CRC has been checked)
        crc = self.crc_func(struct.pack('<B', TYPE_START_APP))
        pkt = struct.pack('<BIB', ord(START_BYTE), crc, TYPE_START_APP)
        self.write(pkt)
        
    def connect(self, reboot_option=False):
        if (self.ser == False) and (self.tx_skt == False):
            return False            

        # Reboot device
        if (reboot_option == "s"):
            self.write(array.array('B', [0xa5, 0x00, 0x00, 0x52, 0x00, 0x66, 0xe3]))
        elif (reboot_option == "o"):
            self.write("\nrestart\n")
        time.sleep(0.1)

        crc = self.crc_func(struct.pack('<B', TYPE_CONNECT))
        pkt = struct.pack('<BIB', ord(START_BYTE), crc, TYPE_CONNECT)
        # Wait for connection with retry
        for j in range(0,CONNECT_RETRIES):
            self.write(pkt)
            while (self.wait_pkt()):
                if (self.type == TYPE_CONNECT_ACK):
                    self.connected = True
                    return True
        return False
                
    def wait_pkt(self):
        start = time.time()
        while (time.time() < (start+COM_TIMEOUT)):
            data = self.read()
            if data:
                for c in data:
                    if (self.state == STATE_IDLE):
                          if (c == START_BYTE):
                            # Start character detected
                            self.state = STATE_HEADER
                            self.pl = ""
                                        
                    elif (self.state == STATE_HEADER):
                        self.pl += c
                        if (len(self.pl) == 5):
                            # End of header reached
                            [self.crc, self.type] = struct.unpack("<IB", self.pl)
                            self.pl = ""
                            self.state = STATE_PAYLOAD
                            if (self.type == TYPE_CONNECT_ACK):
                                self.len = 22
                            elif (self.type == TYPE_WRITE_ROW_ACK):
                                self.len = 3
                            elif (self.type == TYPE_ERASE_PAGE_ACK):
                                self.len = 3
                            else:
                                # Unknown type
                                self.state = STATE_IDLE

                    elif (self.state == STATE_PAYLOAD):
                        self.pl += c
                        if (len(self.pl) == self.len):
                            # Target payload length reached
                            self.state = STATE_IDLE
                            # Check CRC
                            if (self.crc == self.crc_func(struct.pack('<B', self.type) + self.pl)):
                                return True
                            else:
                                # CRC fail
                                pass
                time.sleep(0.05)
        return False

    def get_connection_data(self):
        [bl_ver, sn, cid, pid, app_ver, app_cid, app_pid, app_size, app_crc] = struct.unpack("<HHHHHHHII", self.pl)
        s =  "BL version     %d.%d build %d\n" % (((bl_ver&0xf000) >> 12), ((bl_ver&0x0f00) >> 8), (bl_ver&0x00ff))
        s += "Serial number  %d\n" % sn
        s += "Company ID     %d\n" % cid
        s += "Product ID     %d\n" % pid
        s += "App version    %d.%d build %d\n" % (((app_ver&0xf000) >> 12), ((app_ver&0x0f00) >> 8), (app_ver&0x00ff))
        s += "App company ID %d\n" % app_cid
        s += "App product ID %d\n" % app_pid
        s += "App size       %d bytes\n" % app_size
        s += "App CRC        0x%02x\n" % app_crc
        return s

    def select_file(self, file_path):
        # Get file size
        try:
            file_size = os.path.getsize(file_path)
        except Exception:
            return False
        if (file_size%ROW_SIZE):
            return False
        self.file_path = file_path
        self.total_rows = file_size/ROW_SIZE
        self.total_pages = file_size/PAGE_SIZE
        if (file_size%ROW_SIZE):
            self.total_rows += 1
        if (file_size%PAGE_SIZE):
            self.total_pages += 1
        return True


def main(argv):
    reboot_type = False
    serial_port = False
    ip_address = False

    b = bootloader()

    # Get arguments
    try:
        opts, args = getopt.getopt(argv,"hf:p:r:i:",["help","file=","port=", "reboot=", "ip="])
    except getopt.GetoptError:
        print 'uploader.py -f <file_name> -p <serial_port> -r <reboot_type> -i <ip_address>'
        print 'Reboot types: "s" = MR Sensor Unit, "o" = Skagmo Electronics Owl'
        sys.exit(2)
    for opt, arg in opts:
        arg = arg.strip()
        if opt == '-h':
            print 'uploader.py -f <file_name> -p <serial_port> -r <reboot_type> -i <ip_address>'
            print 'Reboot types: "s" = MR Sensor Unit, "o" = Skagmo Electronics Owl'
            sys.exit()
        elif opt in ("-f"):
            file_name = arg
        elif opt in ("-p"):
            serial_port = arg
        elif opt in ("-r"):
            reboot_type = arg
        elif opt in ("-i"):
            ip_address = arg

    if not b.select_file(file_name):
        print 'Check file.'
        sys.exit()

    if serial_port:
        if not b.serial_open(serial_port, SERIAL_BAUD):
            print 'Check serial port.'
            sys.exit()
    elif ip_address:
        if not b.socket_open(ip_address):
            print 'Check IP address.'
            sys.exit()
    else:
        print 'No serial port or IP defined.'
        sys.exit()

    # Connect to device
    print "Attempting to connect..."
    if (b.connect(reboot_type)):
        print b.get_connection_data()
    else:
        "Connection failed."
        sys.exit()

    # Erase old firmware
    print "Erasing..."
    for page in range (0, b.total_pages):     
        if not b.erase_page(page):
            print "Failed to erase page %d." % page
            sys.exit()

    # Program new firmware
    with open(file_name, 'rb') as f:
        for row in range (0, b.total_rows):       
            print ("Writing row %d/%d (this may take a few minutes)...\r" % (row, b.total_rows-1)),
            sys.stdout.flush()
            if not b.write_row(row, f.read(ROW_SIZE)):
                print "Failed to write row %d\r." % row
                sys.exit()

    # Start new application
    print ""
    print "Successfully wrote new firmware. Booting to new app."
    b.start_app()
    b.close()

if __name__ == "__main__":
   main(sys.argv[1:])

